from lib2to3.pgen2.token import OP
from typing import Optional

import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin

from holisticai.utils.transformers.bias import BMInprocessing as BMImp

from .algorithm import PrejudiceRemoverAlgorithm
from .algorithm_utils import ObjectiveFunction, PRLogger
from .losses import PRBinaryCrossEntropy
from .model import PRLogiticRegression, PRParamInitializer


class PrejudiceRemover(BaseEstimator, ClassifierMixin, BMImp):
    """
    Prejudice remover is an in-processing technique that adds a
    discrimination-aware regularization term to the learning objective.

    References
    ----------
        Kamishima, Toshihiro, et al. "Fairness-aware classifier with prejudice remover regularizer."
        Joint European conference on machine learning and knowledge discovery in databases.
        Springer, Berlin, Heidelberg, 2012.
    """

    def __init__(
        self,
        eta: Optional[float] = 1.0,
        C: Optional[float] = 1.0,
        fit_intercept: Optional[bool] = True,
        penalty: Optional[str] = "l2",
        init_type: Optional[str] = "Zero",
        maxiter: Optional[int] = 1000,
        verbose: Optional[int] = 0,
        print_interval: Optional[int] = 20,
    ):
        """
        Parameters
        ----------
            eta : float
                fairness penalty parameter

            C : float
                Inverse of regularization strength (same as sklearn).

            fit_intercept : bool
                Specifies if a constant must be added to the decision function (same as sklearn).

            penalty : str
                Specify the norm of the penalty (same as sklearn).

            init_type : str
                Specifies how the model parameters will be initialized:
                - Zero : Set all model parameters with zero value.
                - Random : Initialize model parameters with random values.
                - StandarLR : Initialize model parameters with fitted sklearn LR.
                - StandarLRbyGroup : Initialize model parameters with fitted sklearn LR for group_a and group_b.

            maxiter : str
                Maximum number of iterations.

            verbose : int
                Log progress if value > 0.

            print_interval : int
                Each `print_interval` steps print information.

        """

        # Default estimator parameters
        self.eta = eta
        self.C = C
        self.fit_intercept = fit_intercept
        self.penalty = penalty
        self.init_type = init_type

        self.maxiter = maxiter
        self.verbose = verbose
        self.print_interval = print_interval

    def transform_estimator(self):
        # Regularized Binary Cross Entropy
        loss_ = PRBinaryCrossEntropy(C=self.C, eta=self.eta)

        # Model Parameter Initializer
        initializer_ = PRParamInitializer(
            init_type=self.init_type,
            C=self.C,
            penalty=self.penalty,
            fit_intercept=self.fit_intercept,
        )

        # Model
        self.estimator = PRLogiticRegression(
            initializer=initializer_, loss=loss_, fit_intercept=self.fit_intercept
        )
        return self

    def _build_algorithm(self):
        # Support class for logger
        logger_ = PRLogger(
            total_iterations=self.maxiter,
            verbose=self.verbose,
            print_interval=self.print_interval,
        )

        # Support class for objective Function
        objective_fn_ = ObjectiveFunction(
            estimator=self.estimator, loss_fn=self.estimator.loss
        )

        # Algorithm class
        algorithm = PrejudiceRemoverAlgorithm(
            estimator=self.estimator,
            logger=logger_,
            objective_fn=objective_fn_,
            maxiter=self.maxiter,
        )

        return algorithm

    def fit(
        self,
        X: np.ndarray,
        y_true: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit the model

        Description
        -----------
        Learns the regularized logistic regression model.

        Parameters
        ----------
            X : ndarray
                input data

            y_true : ndarray
                Target vector

            group_a : ndarray
                group mask

            group_b : ndarray
                group mask
        Returns:
            Self
        """
        params = self._load_data(X=X, y_true=y_true, group_a=group_a, group_b=group_b)
        X = params["X"]
        y_true = params["y_true"]
        group_b = params["group_b"]
        group_a = params["group_a"]
        sensitive_features = np.stack([group_a, group_b], axis=1)

        self.classes_ = params["classes_"]

        self.algorithm = self._build_algorithm()
        self.algorithm.fit(X=X, y=y_true, sensitive_features=sensitive_features)
        return self

    def predict(self, X: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Prediction

        Description
        ----------
        Predict output for the given samples.

        Parameters
        ----------
        X : pandas.DataFrame or numpy array
            Test samples.

        group_a : ndarray
                group mask

        group_b : ndarray
                group mask

        Returns
        -------

        numpy.ndarray: Predicted output per sample.
        """
        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        X = params["X"]
        group_b = params["group_b"]
        group_a = params["group_a"]
        sensitive_features = np.stack([group_a, group_b], axis=1)
        return self.algorithm.predict(X, sensitive_features)

    def predict_proba(self, X: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Prediction

        Description
        ----------
        Predict output for the given samples.

        Parameters
        ----------
        X : numpy array
            Test samples.

        group_a : ndarray
                group mask

        group_b : ndarray
                group mask

        Returns
        -------

        numpy.ndarray: Predicted output per sample.
        """
        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        X = params["X"]
        group_b = params["group_b"]
        group_a = params["group_a"]
        sensitive_features = np.stack([group_a, group_b], axis=1)
        return self.algorithm.predict_proba(X, sensitive_features)



import numpy as np


class SquareLoss:
    """Class to evaluate the square loss.
    Read more in the :ref:`User Guide <constraints_regression>`.
    """

    def __init__(self, min_val, max_val):
        self.min_val = min_val
        self.max_val = max_val
        self.min = 0
        self.max = (max_val - min_val) ** 2

    def eval(self, y_true, y_pred):  # noqa: A003
        """Evaluate the square loss for the given set of true and predicted values."""
        y_true = np.clip(y_true, self.min_val, self.max_val)
        y_pred = np.clip(y_pred, self.min_val, self.max_val)
        return (y_true - y_pred) ** 2


class AbsoluteLoss:
    """Class to evaluate absolute loss.
    Read more in the :ref:`User Guide <constraints_regression>`.
    """

    def __init__(self, min_val, max_val):
        self.min_val = min_val
        self.max_val = max_val
        self.min = 0
        self.max = np.abs(max_val - min_val)

    def eval(self, y_true, y_pred):  # noqa: A003
        """Evaluate the absolute loss for the given set of true and predicted values."""
        y_true = np.clip(y_true, self.min_val, self.max_val)
        y_pred = np.clip(y_pred, self.min_val, self.max_val)
        return np.abs(y_true - y_pred)


class ZeroOneLoss(AbsoluteLoss):
    """Class to evaluate a zero-one loss.
    Read more in the :ref:`User Guide <constraints_regression>`.
    """

    def __init__(self):
        super().__init__(0, 1)



# Base Imports
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

# sklearn imports
from sklearn.metrics import roc_curve

# utils
from ...utils import get_colors
from ...utils._validation import _check_binary, _regression_checks


def abroca_plot(group_a, group_b, y_score, y_true, ax=None, size=None, title=None):
    """
    Abroca plot

    Description
    -----------
    This function plots the roc curve for both groups
    revealing the area between them (abroca).

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_score : array-like
        Probability estimates (regression)
    y_true : array-like
        Target vector (binary)
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # check and coerce
    group_a, group_b, y_score, y_true, _ = _regression_checks(
        group_a, group_b, y_score, y_true, None
    )
    _check_binary(y_true, "y_true")

    # split data by groups
    y_true_a = y_true[group_a == 1]
    y_score_a = y_score[group_a == 1]
    y_true_b = y_true[group_b == 1]
    y_score_b = y_score[group_b == 1]
    fpr_a, tpr_a, _ = roc_curve(y_true_a, y_score_a, pos_label=1)
    fpr_b, tpr_b, _ = roc_curve(y_true_b, y_score_b, pos_label=1)

    # setup
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("Abroca plot")

    # charting
    colors = get_colors(2)
    plt.plot(fpr_a, tpr_a, label="roc curve group a", color=colors[0])
    plt.plot(fpr_b, tpr_b, label="roc curve group b", color=colors[1])
    plt.fill(
        np.append(fpr_a, fpr_b[::-1]),
        np.append(tpr_a, tpr_b[::-1]),
        color="grey",
        alpha=0.3,
    )
    ax.set_xlabel("fpr")
    ax.set_ylabel("tpr")
    ax.legend()

    # return
    return ax



from xml.dom import NotSupportedErr

from holisticai.utils.transformers.bias import (
    BIAS_TAGS,
    BMInprocessing,
    BMPostprocessing,
    BMPreprocessing,
)


class UTransformersHandler:
    def __init__(self, steps, params_hdl, estimator_hdl):
        """
        Initialize step groups and apply some validations.

        Description
        ----------
        Create bias mitigation groups for preprocessing, inprocessing and postprocessing strategies.
        Pipeline support only one postprocessing in the pipeline.

        Parameters
        ----------
        steps : list
            Pipeline steps

        params_hdl : UTransformersHandler
            Pipeline parameters handler during fit, fit_transform, transform function execution.
        """
        self.bias_mitigators_validation(steps)
        self.steps_groups = {
            tag: [step for step in steps if step[0].startswith(tag)]
            for tag in BIAS_TAGS
        }
        for tag, steps in self.steps_groups.items():
            for step in steps:
                step[1].link_parameters(params_hdl, estimator_hdl)

    def bias_mitigators_validation(self, steps):
        """Validate stem words and bias mitigator position in the pipeline"""
        tag2info = {
            BIAS_TAGS.PRE: BMPreprocessing,
            BIAS_TAGS.INP: BMInprocessing,
            BIAS_TAGS.POST: BMPostprocessing,
        }

        mitigator_groups_by_name = {
            tag: [step for step in steps if step[0].startswith(tag)]
            for tag in BIAS_TAGS
        }
        mitigator_groups_by_object = {
            tag: [step for step in steps if isinstance(step[1], tag2info[tag])]
            for tag in BIAS_TAGS
        }

        # Validate if all objects with bias mitigator stem (BIAS_TAGS) are linked with a correct Bias Mitigator Transformer objects
        if not mitigator_groups_by_name == mitigator_groups_by_object:
            raise NameError(
                f"Mitigator objects and name doesn't match, grouped by name: {mitigator_groups_by_name} \
                and grouped by object type:{mitigator_groups_by_object}"
            )

        num_post_mitigators = len(mitigator_groups_by_name[BIAS_TAGS.POST])

        # Validate if postprocessor bias mitigators are defined after classifier
        if num_post_mitigators > 0:
            post_classifier_step_names, _ = zip(
                *mitigator_groups_by_name[BIAS_TAGS.POST]
            )
            assert all(
                [name.startswith(BIAS_TAGS.POST) for name in post_classifier_step_names]
            ), f"Only bias mitigators postprocessor are supported, \
                utransformer postprocessors founded: {post_classifier_step_names}"

        # Validate that exists only one bias mitigator postprocessor
        # TODO: Evaluate in other cases.
        if not len(mitigator_groups_by_name[BIAS_TAGS.POST]) <= 1:
            raise NotSupportedErr("Pipeline supports max 1 postprocessor mitigator.")

    def drop_post_processing_steps(self, steps):
        """
        Drop post-processing steps from input steps list.

        Description
        ----------
        The function check the steps names and drop names with prefix 'bm_pos'.

        Parameters
        ----------
        steps : list
            Pipeline steps

        Returns
        -------
        list
            New steps list
        """
        return [step for step in steps if not step[0].startswith(BIAS_TAGS.POST)]

    @property
    def pos_processing_steps(self):
        """
        Return bias mitigation post-estimator steps.
        """
        return self.steps_groups[BIAS_TAGS.POST]

    def fit_postprocessing(self, Xt, y, **kargs):
        """
        Fit the pos-estimator transformers.

        Description
        ----------
        Call `fit` of each post-estimator transformer in the pipeline.

        Parameters
        ----------
        Xt : numpy array
            Transformer input data

        y: numpy array
            Target vector

        Returns
        -------
        None
        """
        step = self.pos_processing_steps[0]
        step[1].fit(X=Xt, y_true=y, **kargs)

    def transform_postprocessing(self, **kargs):
        """
        Compute the transformed prediction vector with pos-estimator transformers.

        Description
        ----------
        Call `transform` of each post-estimator transformer in the pipeline.

        Parameters
        ----------
        y_pred : list
            predicted label vector

        y_proba : list
            predicted probability vector

        Returns
        -------
        dict
            Dictionaty with post-processed prediction vectors
        """
        step = self.pos_processing_steps[0]
        yt = step[1].transform(**kargs)
        return yt



import numpy as np
import scipy as sp
from sklearn.metrics import roc_curve

from ..binary_balancer import algorithm_utils as tools


class BinaryBalancerAlgorithm:
    def __init__(self, constraint, objective, threshold_objective="j", binom=False):
        self.constraint = constraint
        self.thr_obj = threshold_objective
        self.objective = objective
        self.binom = binom

    def fit(self, y_true, p_attr, y_pred=None, y_proba=None):

        assert not (
            (y_pred is None) and (y_proba is None)
        ), f"y_pred or y_proba must be passed"

        # Getting the group info
        self.groups = np.unique(p_attr)
        group_ids = [np.where(p_attr == g)[0] for g in self.groups]

        # Optionally thresholding probabilities to get class predictions
        if y_proba is not None:
            probs = y_proba[:, 1]
            self.rocs = [roc_curve(y_true[ids], probs[ids]) for ids in group_ids]
            self.__roc_stats = [
                tools.loss_from_roc(y_true[ids], probs[ids], self.rocs[i])
                for i, ids in enumerate(group_ids)
            ]
            if self.thr_obj == "j":
                cut_ids = [np.argmax(rs["js"]) for rs in self.__roc_stats]
                self.cuts = [self.rocs[i][2][id] for i, id in enumerate(cut_ids)]
                for g, cut in enumerate(self.cuts):
                    probs[group_ids[g]] = tools.threshold(probs[group_ids[g]], cut)
                y_pred = probs.astype(np.uint8)

        # Calcuating the groupwise classification rates
        rates_by_group = [tools.CLFRates(y_true[i], y_pred[i]) for i in group_ids]
        self.group_rates = dict(zip(self.groups, rates_by_group))

        # And then the overall rates
        self.overall_rates = tools.CLFRates(y_true, y_pred)

        # Getting the coefficients for the objective
        p = [len(cols) / len(y_true) for cols in group_ids]

        dr = [(g.nr * p[i], g.pr * p[i]) for i, g in enumerate(rates_by_group)]

        obj_coefs = self.objective(dr, self.overall_rates)

        con, con_bounds = self.constraint.get_conditions(self.group_rates, self.groups)

        # Running the optimization
        self.opt = sp.optimize.linprog(
            c=obj_coefs, bounds=[(0, 1)], A_eq=con, b_eq=con_bounds, method="highs"
        )

        self.pya = self.opt.x.reshape(len(self.groups), 2)

        # Setting the adjusted predictions
        self.y_adj = tools.pred_from_pya(
            y_pred=y_pred, p_attr=p_attr, pya=self.pya, binom=self.binom
        )

    def predict(self, p_attr, y_pred=None, y_proba=None, binom=False):
        """Generates bias-adjusted predictions on new data.
        
        Parameters
        ----------
        y_ : ndarry of shape (n_samples,)
            A binary- or real-valued array of unadjusted predictions.
        
        a : ndarray of shape (n_samples,)
            The protected attributes for the samples in y_.
        
        binom : bool, default False
            Whether to generate adjusted predictions by sampling from a \
            binomial distribution.
        
        Returns
        -------
        y~ : ndarray of shape (n_samples,)
            The adjusted binary predictions.
        """
        # Optional thresholding for continuous predictors
        if y_proba is not None:
            probs = y_proba[:, 1]
            group_ids = [np.where(p_attr == g)[0] for g in self.groups]
            for g, cut in enumerate(self.cuts):
                y_pred[group_ids[g]] = tools.threshold(probs[group_ids[g]], cut)

        # Returning the adjusted predictions
        adj = tools.pred_from_pya(y_pred, p_attr, self.pya, binom)
        return adj



# Base Imports
from os import environ, makedirs
from os.path import expanduser, join

# sklearn imports
from sklearn.datasets import fetch_openml


def get_data_home(data_home=None):
    """
    Return the path of the holisticai data directory.
    By default the data directory is set to a folder named 'holisticai_data' in the
    user home folder.
    Alternatively, it can be set by the 'HOLISTICAI_DATA' environment
    variable or programmatically by giving an explicit folder path. The '~'
    symbol is expanded to the user home folder.
    If the folder does not already exist, it is automatically created.
    Parameters
    ----------
    data_home : str, default=None
        The path to holisticai data directory. If `None`, the default path
        is `~/holisticai_data`.
    Returns
    -------
    data_home: str
        The path to holisticai data directory.
    """
    if data_home is None:
        data_home = environ.get("HOLISTIC_AI_DATA", join("~", "holisticai_data"))
    data_home = expanduser(data_home)
    makedirs(data_home, exist_ok=True)
    return data_home


def load_student(data_home=None, return_X_y=False, as_frame=True):
    """
    Load Student Dataset.

    Description
    ----------
    We use the sklearn.datasets fetch_openml function to
    fetch data from the openml api. Information about
    the Student dataset can be obtained by printing
    load_student()['DESCR'].

    Parameters
    ----------
    data_home : str
        The directory to which the data is downloaded. If None, we dowload
        to the holisticai_data folder in user home directory.
    return_X_y : bool, default=False
        If True, returns ``(data, target)`` instead of a Bunch object.
        See below for more information about the `data` and `target` object.
    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with
        appropriate dtypes (numeric). The target is
        a pandas DataFrame or Series depending on the number of target columns.
        If `return_X_y` is True, then (`data`, `target`) will be pandas
        DataFrames or Series as described below.

    Returns
    -------
    data : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.
        data : {ndarray, dataframe} of shape (1797, 64)
            The flattened data matrix. If `as_frame=True`, `data` will be
            a pandas DataFrame.
        target: {ndarray, Series} of shape (1797,)
            The classification target. If `as_frame=True`, `target` will be
            a pandas Series.
        frame: DataFrame of shape (1797, 65)
            Only present when `as_frame=True`. DataFrame with `data` and
            `target`.
        feature_names: list
            The names of the dataset columns.
        target_names: list
            The names of target classes.
        DESCR: str
            The full description of the dataset.
        (data, target) : tuple if ``return_X_y`` is True
            A tuple of two ndarrays by default. The first contains a 2D ndarray of
            shape (num_rows, num_columns) with each row representing one sample and
            each column representing the features. The second ndarray of shape (1797)
            contains the target samples.  If `as_frame=True`, both arrays are pandas
            objects, i.e. `X` a dataframe and `y` a series.
    """
    if data_home is None:
        data_home = get_data_home()

    bunch = fetch_openml(
        name="UCI-student-performance-mat",
        version=1,
        data_home=data_home,
        return_X_y=return_X_y,
        as_frame=as_frame,
    )

    return bunch


def load_adult(data_home=None, return_X_y=False, as_frame=True):
    """
    Load Adult Dataset.

    Description
    ----------
    We use the sklearn.datasets fetch_openml function to
    fetch data from the openml api. Information about
    the Adult dataset can be obtained by printing
    load_adult()['DESCR'].

    Parameters
    ----------
    data_home : str
        The directory to which the data is downloaded. If None, we download
        to the holisticai_data folder in user home directory.
    return_X_y : bool, default=False
        If True, returns ``(data, target)`` instead of a Bunch object.
        See below for more information about the `data` and `target` object.
    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with
        appropriate dtypes (numeric). The target is
        a pandas DataFrame or Series depending on the number of target columns.
        If `return_X_y` is True, then (`data`, `target`) will be pandas
        DataFrames or Series as described below.

    Returns
    -------
    data : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.
        data : {ndarray, dataframe} of shape (1797, 64)
            The flattened data matrix. If `as_frame=True`, `data` will be
            a pandas DataFrame.
        target: {ndarray, Series} of shape (1797,)
            The classification target. If `as_frame=True`, `target` will be
            a pandas Series.
        frame: DataFrame of shape (1797, 65)
            Only present when `as_frame=True`. DataFrame with `data` and
            `target`.
        feature_names: list
            The names of the dataset columns.
        target_names: list
            The names of target classes.
        DESCR: str
            The full description of the dataset.
        (data, target) : tuple if ``return_X_y`` is True
            A tuple of two ndarrays by default. The first contains a 2D ndarray of
            shape (num_rows, num_columns) with each row representing one sample and
            each column representing the features. The second ndarray of shape (1797)
            contains the target samples.  If `as_frame=True`, both arrays are pandas
            objects, i.e. `X` a dataframe and `y` a series.
    """
    if data_home is None:
        data_home = get_data_home()

    bunch = fetch_openml(
        name="adult",
        version=2,
        data_home=data_home,
        return_X_y=return_X_y,
        as_frame=as_frame,
    )

    return bunch


def load_law_school(data_home=None, return_X_y=False, as_frame=True):
    """
    Load Law School Dataset.

    Description
    ----------
    We use the sklearn.datasets fetch_openml function to
    fetch data from the openml api. Information about
    the Law School dataset can be obtained by printing
    load_law_school()['DESCR'].

    Parameters
    ----------
    data_home : str
        The directory to which the data is downloaded. If None, we download
        to the holisticai_data folder in user home directory.
    return_X_y : bool, default=False
        If True, returns ``(data, target)`` instead of a Bunch object.
        See below for more information about the `data` and `target` object.
    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with
        appropriate dtypes (numeric). The target is
        a pandas DataFrame or Series depending on the number of target columns.
        If `return_X_y` is True, then (`data`, `target`) will be pandas
        DataFrames or Series as described below.

    Returns
    -------
    data : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.
        data : {ndarray, dataframe} of shape (1797, 64)
            The flattened data matrix. If `as_frame=True`, `data` will be
            a pandas DataFrame.
        target: {ndarray, Series} of shape (1797,)
            The classification target. If `as_frame=True`, `target` will be
            a pandas Series.
        frame: DataFrame of shape (1797, 65)
            Only present when `as_frame=True`. DataFrame with `data` and
            `target`.
        feature_names: list
            The names of the dataset columns.
        target_names: list
            The names of target classes.
        DESCR: str
            The full description of the dataset.
        (data, target) : tuple if ``return_X_y`` is True
            A tuple of two ndarrays by default. The first contains a 2D ndarray of
            shape (num_rows, num_columns) with each row representing one sample and
            each column representing the features. The second ndarray of shape (1797)
            contains the target samples.  If `as_frame=True`, both arrays are pandas
            objects, i.e. `X` a dataframe and `y` a series.
    """
    if data_home is None:
        data_home = get_data_home()

    bunch = fetch_openml(
        name="law-school-admission-bianry",
        version=1,
        data_home=data_home,
        return_X_y=return_X_y,
        as_frame=as_frame,
    )

    return bunch


def load_last_fm(data_home=None, return_X_y=False, as_frame=True):
    """
    Load Last FM Dataset.

    Description
    ----------
    We use the sklearn.datasets fetch_openml function to
    fetch data from the openml api. Information about
    the Last FM dataset can be obtained by printing
    load_last_fm()['DESCR'].

    Parameters
    ----------
    data_home : str
        The directory to which the data is downloaded. If None, we download
        to the holisticai_data folder in user home directory.
    return_X_y : bool, default=False
        If True, returns ``(data, target)`` instead of a Bunch object.
        See below for more information about the `data` and `target` object.
    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with
        appropriate dtypes (numeric). The target is
        a pandas DataFrame or Series depending on the number of target columns.
        If `return_X_y` is True, then (`data`, `target`) will be pandas
        DataFrames or Series as described below.

    Returns
    -------
    data : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.
        data : {ndarray, dataframe} of shape (1797, 64)
            The flattened data matrix. If `as_frame=True`, `data` will be
            a pandas DataFrame.
        target: {ndarray, Series} of shape (1797,)
            The classification target. If `as_frame=True`, `target` will be
            a pandas Series.
        frame: DataFrame of shape (1797, 65)
            Only present when `as_frame=True`. DataFrame with `data` and
            `target`.
        feature_names: list
            The names of the dataset columns.
        target_names: list
            The names of target classes.
        DESCR: str
            The full description of the dataset.
        (data, target) : tuple if ``return_X_y`` is True
            A tuple of two ndarrays by default. The first contains a 2D ndarray of
            shape (num_rows, num_columns) with each row representing one sample and
            each column representing the features. The second ndarray of shape (1797)
            contains the target samples.  If `as_frame=True`, both arrays are pandas
            objects, i.e. `X` a dataframe and `y` a series.
    """
    if data_home is None:
        data_home = get_data_home()

    bunch = fetch_openml(
        name="LastFM_dataset",
        version=1,
        data_home=data_home,
        return_X_y=return_X_y,
        as_frame=as_frame,
    )

    return bunch


def load_us_crime(data_home=None, return_X_y=False, as_frame=True):
    """
    Load US Crime Dataset.

    Description
    ----------
    We use the sklearn.datasets fetch_openml function to
    fetch data from the openml api. Information about
    the US Crime dataset can be obtained by printing
    load_us_crime()['DESCR'].

    Parameters
    ----------
    data_home : str
        The directory to which the data is downloaded. If None, we download
        to the holisticai_data folder in user home directory.
    return_X_y : bool, default=False
        If True, returns ``(data, target)`` instead of a Bunch object.
        See below for more information about the `data` and `target` object.
    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with
        appropriate dtypes (numeric). The target is
        a pandas DataFrame or Series depending on the number of target columns.
        If `return_X_y` is True, then (`data`, `target`) will be pandas
        DataFrames or Series as described below.

    Returns
    -------
    data : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.
        data : {ndarray, dataframe} of shape (1994, 127)
            The flattened data matrix. If `as_frame=True`, `data` will be
            a pandas DataFrame.
        target: {ndarray, Series} of shape (1994,)
            The classification target. If `as_frame=True`, `target` will be
            a pandas Series.
        frame: DataFrame of shape (1994, 128)
            Only present when `as_frame=True`. DataFrame with `data` and
            `target`.
        feature_names: list
            The names of the dataset columns.
        target_names: list
            The names of target classes.
        DESCR: str
            The full description of the dataset.
        (data, target) : tuple if ``return_X_y`` is True
            A tuple of two ndarrays by default. The first contains a 2D ndarray of
            shape (num_rows, num_columns) with each row representing one sample and
            each column representing the features. The second ndarray of shape (1994)
            contains the target samples.  If `as_frame=True`, both arrays are pandas
            objects, i.e. `X` a dataframe and `y` a series.
    """
    if data_home is None:
        data_home = get_data_home()

    bunch = fetch_openml(
        name="us_crime",
        version=1,
        data_home=data_home,
        return_X_y=return_X_y,
        as_frame=as_frame,
    )

    return bunch



from functools import partial

import numpy as np
from scipy.stats import multivariate_normal
from sklearn.metrics import accuracy_score

from holisticai.utils.transformers.bias import SensitiveGroups


def prob(dist, x):
    return dist.pdf(x)


class MetaFairClassifierAlgorithm:
    def __init__(self, logger, tau, eps=0.01, batch_size=20, constraint=None, steps=10):
        self.tau = tau
        self.eps = eps
        self.batch_size = batch_size
        self.steps = steps
        self.constraint = constraint
        self.logger = logger
        self.sens_groups = SensitiveGroups()

    def range(self, eps, tau):
        a = np.arange(np.ceil(tau / eps), step=self.steps) * eps
        b = (a + eps) / tau
        b = np.minimum(b, 1)
        return np.c_[a, b]

    def probabilities(self, dist, x):
        pos = np.ones(len(x))
        zero = np.zeros(len(x))

        P = {}

        P["Y=1,Z=1"] = prob(dist, np.c_[x, pos, pos])
        P["Y=-1,Z=1"] = prob(dist, np.c_[x, -pos, pos])
        P["Y=1,Z=0"] = prob(dist, np.c_[x, pos, zero])
        P["Y=-1,Z=0"] = prob(dist, np.c_[x, -pos, zero])

        P["total"] = P["Y=1,Z=1"] + P["Y=1,Z=0"] + P["Y=-1,Z=0"] + P["Y=-1,Z=1"]

        P["Y=1"] = (P["Y=1,Z=1"] + P["Y=1,Z=0"]) / P["total"]
        P["Z=0"] = (P["Y=-1,Z=0"] + P["Y=1,Z=0"]) / P["total"]
        P["Z=1"] = (P["Y=-1,Z=1"] + P["Y=1,Z=1"]) / P["total"]

        return P

    def compute_gradient(self, P, a, b):
        """Gradient Descent implementation for the optimizing the objective
        function.

        Note that one can alternately also use packages like CVXPY here.
        Here we use decaying step size. For certain objectives, constant step
        size might be better.
        """

        def init_params(i):
            return [i - 5] * self.constraint.num_params

        min_cost = np.inf  # 1e8
        best_param = None
        for i in range(1, 10):
            params = init_params(i)
            for k in range(1, 50):
                grad = self.constraint.expected_gradient(P, params, a, b)
                for j in range(self.constraint.num_params):
                    params[j] = params[j] - 1 / k * grad[j]

                cost = self.constraint.cost_function(P, params, a, b)
                # print(cost)
                if cost < min_cost:
                    min_cost, best_param = cost, params
        return best_param

    def forward(self, dist, params, a, b, x):
        P = self.probabilities(dist=dist, x=x)
        return self.constraint.forward(P=P, params=params, a=a, b=b)

    def fit(self, X, y, sensitive_features, random_state=None):
        y_true = y.copy()
        y_true[y == 0] = -1

        groups_num = self.sens_groups.fit_transform(
            sensitive_features, convert_numeric=True
        )

        """Returns the model given the training data and input tau."""
        train = np.concatenate(
            [X, y_true[:, np.newaxis], groups_num[:, np.newaxis]], axis=1
        )
        mean = np.mean(train, axis=0)
        cov = np.cov(train, rowvar=False)
        dist = multivariate_normal(mean, cov, allow_singular=True, seed=random_state)
        nb_features = X.shape[1]
        dist_x = multivariate_normal(
            mean[:nb_features],
            cov[:nb_features, :nb_features],
            allow_singular=True,
            seed=random_state,
        )

        z_prior = np.mean(groups_num)
        self.constraint.z_prior = z_prior
        max_acc = -np.inf
        params_opt = [0] * self.constraint.num_params
        p, q = 0, 0
        for it, (a, b) in enumerate(self.range(self.eps, self.tau), 1):
            samples = dist_x.rvs(size=self.batch_size)

            P = self.probabilities(dist, samples)

            params = self.compute_gradient(P, a, b)

            t = self.forward(dist, params, a, b, X)

            y_pred = np.where(t > 0, 1, -1)

            acc = accuracy_score(y_true, y_pred)

            gamma = self.constraint.gamma(y_true, y_pred, groups_num)

            if max_acc < acc and gamma >= self.tau - 0.2:  # TODO: why - 0.2?
                max_acc = acc
                params_opt = params
                p, q = a, b

            self.logger.update(it, max_acc, gamma)

        self.predictor = partial(self.forward, dist, params_opt, p, q)
        return self

    def predict(self, X: np.ndarray):
        """
        Prediction

        Description
        ----------
        Predict output for the given samples.

        Parameters
        ----------
        X : pandas.DataFrame or numpy array
            Test samples.

        Returns
        -------

        numpy.ndarray: Predicted output per sample.
        """

        t = self.predictor(X)
        y = (t > 0).astype(int).reshape((-1, 1))
        return y.ravel()

    def predict_proba(self, X: np.ndarray):
        """
        Probability Prediction

        Description
        ----------
        Probability estimate for the given samples.

        Parameters
        ----------
        X : pandas.DataFrame or numpy array
            Test samples.

        Returns
        -------
        numpy.ndarray
            probability output per sample.
        """
        t = self.predictor(X)
        scores = ((t + 1) / 2).reshape((-1, 1))
        return scores



from collections import namedtuple

from holisticai.utils.transformers.bias._group_utils import SensitiveGroups
from holisticai.utils.transformers.bias._inprocessing import BMInprocessing
from holisticai.utils.transformers.bias._postprocessing import BMPostprocessing
from holisticai.utils.transformers.bias._preprocessing import BMPreprocessing

BiasMitigationTags = namedtuple("BiasMitigationTags", ["PRE", "INP", "POST"])
BIAS_TAGS = BiasMitigationTags(PRE="bm_pre", INP="bm_inp", POST="bm_pos")



import numpy as np
from sklearn.metrics.pairwise import pairwise_distances, pairwise_distances_argmin

from holisticai.bias.mitigation.commons.fairlet_clustering._utils import distance


class FairletClusteringAlgorithm:
    def __init__(self, decomposition, clustering_model):
        self.decomposition = decomposition
        self.clustering_model = clustering_model

    def fit(self, X, group_a, group_b, decompose=None):
        if decompose is not None:
            fairlets, fairlet_centers, fairlet_costs = decompose
        else:
            fairlets, fairlet_centers, fairlet_costs = self.decomposition.fit_transform(
                X, group_a, group_b
            )

        self.clustering_model.fit([X[i] for i in fairlet_centers])
        mapping = self.clustering_model.assign()

        self.labels = np.zeros(len(X), dtype="int32")
        for fairlet_id, final_cluster in mapping:
            self.labels[fairlets[fairlet_id]] = int(fairlet_centers[final_cluster])
        self.centers = np.array(
            [fairlet_centers[i] for i in self.clustering_model.centers]
        )
        self.cluster_centers_ = X[self.centers]
        self.cost = np.max(np.min(pairwise_distances(X, Y=X[self.centers]), axis=1))
        self.X = X

    def predict(self, X):
        """
        Decsription
        -----------
        Assigning every point in the dataset to the closest center.

        Paramters
        ---------
            X: matrix-like
            input dataset.

        Returns
        -------
            mapping : list of centers
        """
        fairlets_midxs = pairwise_distances_argmin(X, Y=self.X)
        return self.labels[fairlets_midxs]



import numpy as np
import pandas as pd

from .algorithm import DELTRAlgorithm
from .algorithm_utils import Standarizer


class DebiasingExposure:
    """
    Disparate Exposure Learning to Rank (DELTR) incorporates a measure of performance and a measure
    of disparate exposure into its loss function. Trains a linear model based on performance and
    fairness for a protected group.

    Reference
    ---------
    [1] Zehlike, Meike, and Carlos Castillo. "Reducing disparate exposure in ranking: A learning to rank
        approach." Proceedings of The Web Conference 2020. 2020.
    """

    def __init__(
        self,
        group_col: str,
        query_col="query_id",
        doc_col="doc_id",
        score_col="judgment",
        feature_cols=[],
        gamma: float = 1.0,
        number_of_iterations=3000,
        learning_rate=0.001,
        lambdaa=0.001,
        init_var=0.01,
        standardize=False,
        verbose=0,
    ):
        """
        Description
        -----------
        Init Disparate Exposure in Learning To Rank class.

        Parameters
        ----------
        group_col: str
            Name of the column in data that contains protected attribute.

        query_col:
            Name of the column in data that contains query ids.

        doc_col:
            List of name of the column in data that contains document ids.

        score_col:
            Name of the column in data that contains judgment values.

        feature_cols:
            Name of the columns in data that contains feature values.

        gamma: float
            Gamma parameter for the cost calculation in the training phase (recommended to be around 1).

        number_of_iterations: int
            Number of iteration in gradient descent (optional).

        learning_rate: float
            Learning rate in gradient descent (optional).

        lambdaa: float
            Regularization constant (optional).

        init_var: float
            Range of values for initialization of weights (optional).

        standardize: bool
            Boolean indicating whether the data should be standardized or not (optional).

        verbose : int
            If > 0, print progress.
        """

        # check if mandatory parameters are present
        if group_col is None:
            raise ValueError(
                "The name of column in data `group_col` must be initialized"
            )
        if gamma is None:
            raise ValueError("The `gamma` parameter must be initialized")

        # initialize the protected_feature index to -1

        # assign mandatory parameters
        self.group_col = group_col
        self.query_col = query_col
        self.doc_col = doc_col
        self.score_col = score_col
        self.feature_cols = feature_cols
        self._gamma = gamma

        # assign optional parameters
        self._number_of_iterations = number_of_iterations
        self._learning_rate = learning_rate
        self._lambda = lambdaa
        self._init_var = init_var
        self._standardize = standardize
        self.verbose = verbose
        self.standarizer = Standarizer(group_col=group_col)
        self.algorithm = DELTRAlgorithm(
            self._gamma,
            self._number_of_iterations,
            self._learning_rate,
            self._lambda,
            self._init_var,
            verbose=verbose,
        )

    def _filter_invalid_examples(self, rankings):
        new_rankings = []
        for q, ranking in rankings.groupby(self.query_col):
            if (ranking[self.group_col].sum() > 0).any():
                new_rankings.append(ranking)
        new_rankings = pd.concat(new_rankings, axis=0).reset_index(drop=True)
        return new_rankings

    def fit(self, rankings: pd.DataFrame):
        """
        Description
        -----------
        Trains a Disparate Exposure model on a given training set.

        Parameters
        ----------
        rankings:  DataFrame

        Return
        ------
            Self
        """

        rankings = self._filter_invalid_examples(rankings)
        if self.feature_cols == []:
            restricted_cols = [self.query_col, self.doc_col, self.score_col]
            self.feature_cols = [
                col
                for col in rankings.columns.to_list()
                if not (col in restricted_cols)
            ]

        # prepare data
        (
            query_ids,
            doc_ids,
            protected_feature,
            feature_matrix,
            training_scores,
        ) = self._prepare_data(rankings, has_judgment=True)

        # standardize data if allowed
        if self._standardize:
            feature_matrix = self.standarizer.fit_transform(feature_matrix)

        self._omega = self.algorithm.fit(
            query_ids, protected_feature, training_scores, feature_matrix
        )

        # return model
        return self

    def transform(self, rankings: pd.DataFrame):
        """
        Description
        -----------
        Train a Disparate Exposure model to rank the prediction set.

        Parameters
        ----------
        rankings:    DataFrame

        Return
        ------
            DataFrame
        """

        if self._omega is None:
            raise SystemError("You need to train a model first!")

        # prepare data
        query_ids, doc_ids, protected_attributes, feature_matrix = self._prepare_data(
            rankings, has_judgment=False
        )

        # standardize data if allowed
        if self._standardize:
            feature_matrix = self.standarizer.transform(feature_matrix)

        # calculate the predictions
        predictions = np.dot(feature_matrix, self._omega)

        # create the resulting data frame
        result = pd.DataFrame(
            {
                self.query_col: query_ids,
                self.doc_col: doc_ids,
                self.group_col: protected_attributes,
                self.score_col: predictions,
            }
        )

        # sort by the score in descending order
        result = result.sort_values([self.score_col], ascending=[0])

        return result

    def _prepare_data(self, data, has_judgment=False):
        """
        Description
        -----------
        Extracts the different columns of the input data.

        Parameters
        ----------
        data: DataFrame

        has_adjudment: bool

        Return
        ------
            Tuple of preprocessed data
        """
        query_ids = data[self.query_col]
        doc_ids = data[self.doc_col]
        protected_attributes = data[self.group_col]  # add 2 for query id and doc id

        feature_matrix = data[self.feature_cols]
        if has_judgment:
            scores = data[self.score_col]
            return query_ids, doc_ids, protected_attributes, feature_matrix, scores
        else:
            return query_ids, doc_ids, protected_attributes, feature_matrix



import copy

import numpy as np

from ..randomized_threshold.algorithm import RandomizedThresholdAlgorithm
from .algorithm_utils import RBLogger


class Reduce2BinaryAlgorithm:
    """
    Debiase multiclass datasets via preprocessing (R2B Algorithm).
    Use Alternating Direction Method of multipliers (ADMM) to decompose the problem into separate
    debiasing tasks of binary labels before they are aggregated.
    """

    def __init__(
        self,
        gamma=1.0,
        eps=0,
        eta=0.5,
        num_classes=2,
        sgd_steps=10,
        full_gradient_epochs=1_000,
        batch_size=256,
        max_iter=100,
        verbose=True,
    ):
        """
        Instantiate object.

        Parameters
        ----------
          gamma: float
            The regularization parameter gamma (for randomization). Set this to
            1 if the goal is to minimize changes to the original probability scores.

          eps: float
            Tolerance parameter for bias >= 0.

          eta: float
            The step size parameters in ADMM.

          num_classes: int
            Number of classes (must be >= 2).

          sgd_steps: int
            Number of minibatch steps in SGD.

          full_gradient_epochs: int
            Number of full gradient descent steps.

          batch_size: int
            Size of minibatches in SGD.

          max_iter: int
            Maximum number of iteration of the ADMM procedure.

          verbose: bool
            If True, display progress.
        """
        if num_classes < 2:
            raise ValueError("Number of classes (must be >= 2).")

        if eps < 0:
            raise ValueError("eps must be non-negative.")

        if gamma <= 0:
            raise ValueError("gamma must be a strictly positive number.")

        self.num_groups = 1
        self.gamma = gamma
        self.eps = eps
        self.eta = eta
        self.num_classes = num_classes
        self.max_iter = max_iter

        # binary debiasers for each label
        self.debiasers = {}
        for k in range(num_classes):
            self.debiasers[k] = RandomizedThresholdAlgorithm(
                gamma=gamma + eta,
                eps=eps,
                sgd_steps=sgd_steps,
                full_gradient_epochs=full_gradient_epochs,
                batch_size=batch_size,
                verbose=False,
            )

        self.logger = RBLogger(verbose=verbose, total_iterations=max_iter)

    def _compute_z(self, h_mat, u_mat):
        # Compute the Z matrix in the R2B algorithm.
        mult_by_ones = np.matmul(
            h_mat + u_mat,
            np.ones(
                self.num_classes,
            ),
        )
        over_k = 1.0 / self.num_classes * (mult_by_ones - np.ones(mult_by_ones.shape))
        j_mat = np.outer(
            over_k,
            np.ones(
                self.num_classes,
            ),
        )
        return h_mat + u_mat - j_mat

    def predict(self, y_prob, p_attr):
        """
        Description
        -----------
        Debias scores w.r.t. the sensitive class in each demographic group.
        In the multiclass setting, we use ADMM to decompose the problem into
        separate debiasing tasks of binary labels before they are aggregated.

        Parameters
        ----------
          y_prob: np.ndarray
            Predicted probability matrix.

          p_attr: An array containing the group id of each instance starting
            from group 0 to group (num_classes-1).

        Returns:
          fair infered probability matrix.
        """

        if len(y_prob.shape) != 2:
            raise ValueError(
                "Original prob scores must be a 2-dimensional array."
                "Use RandomizedThreshold for binary classification."
            )

        y_prob_scores = copy.deepcopy(y_prob)

        # Initialize ADMM.
        f_mat = copy.deepcopy(y_prob_scores)
        h_mat = np.zeros_like(f_mat)
        u_mat = np.zeros_like(f_mat)
        z_mat = np.zeros_like(f_mat)

        for iteration in range(self.max_iter):
            # Step 1: debias each label separately.
            for k in range(self.num_classes):
                self.debiasers[k].fit(f_mat[:, k], p_attr)

                h_mat[:, k] = self.debiasers[k].predict(f_mat[:, k], p_attr)

                # Step 2: update ADMM variables.
                old_z = copy.deepcopy(z_mat)
                z_mat = self._compute_z(h_mat, u_mat)
                u_mat = u_mat + h_mat - z_mat
                f_mat = y_prob + self.eta * (z_mat - u_mat)

                # Compute primal and dual residuals.
                s = np.linalg.norm(z_mat - old_z)
                r = np.linalg.norm(z_mat - h_mat)

            self.logger.update(
                iteration=iteration + 1, primal_residual=r, dual_residual=s
            )

        z_mat = np.maximum(z_mat, 0)
        z_mat = z_mat / np.sum(z_mat, axis=1, keepdims=True)
        return z_mat



import numpy as np
import pandas as pd
import scipy

from . import algorithm_utils as tools


class MulticlassBalancerAlgorithm:
    def __init__(self, constraint, objective):
        """
        Initializes an instance of a PredictionBalancer.
        """
        self.constraint = constraint
        self.objective = objective

    def fit(self, y_true, y_pred, p_attr):
        """
        Parameters
        ----------
        y_true : array-like of shape (n_samples,) or str
            The true labels, either as an array or as a string \
            specifying a column in data. 
        
        y_pred : array-like of shape (n_samples,) or str
            The predicted labels, either as an array or as a string \
            specifying a column in data.
        
        p_attr : array-like of shape (n_samples,) or str
            The protected attribute, either as an array, or as a string \
            specifying the column in data.
        """

        assert y_pred is not None, f"y_pred must be passed"

        # Getting the group info
        p_y = tools.p_vec(y_true)
        p_a = tools.p_vec(p_attr)
        self.n_classes = p_y.shape[0]

        # Getting some basic info for each group
        self.groups = np.unique(p_attr)
        self.n_groups = len(self.groups)
        group_ids = [np.where(p_attr == g)[0] for g in self.groups]

        # Getting the group-specific P(Y), P(Y- | Y), and constraint matrices
        p_vecs = np.array([tools.p_vec(y_true[ids]) for ids in group_ids])
        p_vecs = p_a.reshape(-1, 1) * p_vecs
        cp_mats = np.array(
            [
                tools.cp_mat(y_true[ids], y_pred[ids], self.n_classes)
                for ids in group_ids
            ]
        )
        cp_mats_t = np.zeros((self.n_classes, self.n_classes, self.n_groups))

        for a in range(self.n_groups):
            cp_mats_t[:, :, a] = cp_mats[a].transpose()

        self.constraint.n_groups = self.n_groups
        self.objective.n_classes = self.n_classes
        self.objective.n_groups = self.n_groups

        con, con_bounds = self.constraint.get_conditions(p_vecs, p_a, cp_mats)
        self.obj = self.objective(cp_mats)

        # Running the optimization
        opt = scipy.optimize.linprog(
            c=self.obj, bounds=[0, 1], A_eq=con, b_eq=con_bounds, method="highs"
        )

        # Getting the Y~ matrices
        m = tools.pars_to_cpmat(opt, n_groups=self.n_groups, n_classes=self.n_classes)

        # Getting the new cp matrices
        self.new_cp_mats = np.array(
            [np.dot(cp_mats[i], m[i]) for i in range(self.n_groups)]
        )

    def predict(self, y_pred, p_attr, seed=2021):
        """Generates bias-adjusted predictions on new data.

        Parameters
        ----------
        y_pred : ndarry of shape (n_samples,)
            A binary- or real-valued array of unadjusted predictions.

        p_attr : ndarray of shape (n_samples,)
            The protected attributes for the samples in y_.

        Returns
        -------
        y~ : ndarray of shape (n_samples,)
            The adjusted binary predictions.
        """
        assert y_pred is not None, f"y_pred must be passed"

        pd.options.mode.chained_assignment = None
        y_tilde = y_pred.copy()
        np.random.seed(seed)
        seeds = np.random.randint(0, 1e6, len(self.groups))
        for g in self.groups:
            for c in range(self.n_classes):
                y_ids = np.where((p_attr == g) & (y_pred == c))[0]
                np.random.seed(seeds[g])
                y_tilde[y_ids] = np.random.choice(
                    a=self.n_classes, p=self.new_cp_mats[g][c], size=len(y_ids)
                )
        return y_tilde



import random

import numpy as np
from sklearn.metrics.pairwise import pairwise_distances
from tqdm import tqdm


class KMediamClusteringAlgorithm:
    def __init__(
        self,
        n_clusters=2,
        init_centers="Random",
        max_iter=1000,
        strategy="LS",
        verbose=False,
    ):
        self.k = n_clusters
        self.init_centers = init_centers
        self.max_iter = max_iter
        self.verbose = verbose
        self.strategy = strategy

    def _compute_new_assigment_and_cost(self, centers):
        cost = self._compute_cost(centers)
        assignment = self._compute_new_assigment(centers)
        return assignment, cost

    def _compute_cost(self, centers):
        centers = np.array(centers, dtype=np.int32)
        group_costs = []
        for gid in self.group_ids:
            group_costs.append(
                np.mean(
                    np.amin(self.distances[np.ix_(self.p_attr == gid, centers)], axis=1)
                )
            )
        cost = np.max(group_costs)
        return cost

    def _compute_new_assigment(self, centers):
        centers = np.array(centers, dtype=np.int32)
        assignment = centers[
            np.argmin(self.distances[np.ix_(np.arange(self.n), centers)], axis=1)
        ]
        return assignment

    def fit(self, X, p_attr):
        self.n = len(X)
        self.group_ids = np.unique(p_attr)
        distances = pairwise_distances(X, Y=X, metric="l1")

        self.distances = distances
        self.p_attr = p_attr

        if self.strategy == "LS":
            self._linear_search(X)

        elif self.strategy == "GA":
            self._genetic_algorithm(X)

    def _linear_search(self, X):

        if self.init_centers == "KMedoids":
            from holisticai.utils.models.cluster import KMedoids

            kmedoids = KMedoids(n_clusters=self.k).fit(X)
            starting_centers = kmedoids.medoid_indices_

        elif self.init_centers == "Random":
            starting_centers = random.sample(range(1, self.n - 1), self.k)

        if starting_centers[1] > starting_centers[0] - 1:
            starting_centers[1] += 1

        chosen_centers = starting_centers
        centers = starting_centers

        assignment, min_cost = self._compute_new_assigment_and_cost(centers)

        flag = 0
        for _ in range(self.max_iter):
            flag = 1
            # Check if any other point is a better center
            r = list(range(self.n))
            random.shuffle(r)
            t = tqdm(r, desc=f"Cost: {min_cost:.4f}", leave=True)
            for c in t:
                for i in range(self.k):
                    centers = list(chosen_centers)
                    # Try replacing i center
                    if not (c in centers):
                        # set new center
                        centers[i] = c
                        # do new assignment calculate new cost
                        assignment, curr_cost = self._compute_new_assigment_and_cost(
                            centers
                        )
                        # change variables if new cost is better
                        if curr_cost < min_cost:
                            min_cost = curr_cost
                            t.set_description(f"Cost: {min_cost:.4f}")
                            t.refresh()  # to show immediately the update
                            chosen_centers = centers
                            chosen_assignment = assignment
                            flag = 0
                            break

            if flag == 1:
                break

        self.labels_ = np.array(chosen_assignment)
        self.cluster_centers_ = X[np.array(chosen_centers)]
        self.centers = np.array(chosen_centers)

    def _genetic_algorithm(self, X):
        optimization_function = lambda x: self._compute_cost(x)

        from .....utils.optimizers import GAHiperparameters, GeneticAlgorithm

        varbound = np.array([[0, len(X) - 1]] * self.k)

        ga_kargs = {
            "max_num_iteration": self.max_iter,
            "population_size": 100,
            "mutation_probability": 0.1,
            "elit_ratio": 0.01,
            "crossover_probability": 0.5,
            "parents_portion": 0.3,
            "crossover_type": "uniform",
            "max_iteration_without_improv": None,
        }

        algorithm_parameters = GAHiperparameters(**ga_kargs)

        optimizer = GeneticAlgorithm(
            function=optimization_function,
            dimension=self.k,
            variable_type="int",
            variable_boundaries=varbound,
            algorithm_parameters=algorithm_parameters,
            verbose=self.verbose,
        )

        optimizer.run()

        chosen_centers = np.array(optimizer.best_variable, dtype=np.int32)
        chosen_assignment = self._compute_new_assigment(chosen_centers)

        self.labels_ = np.array(chosen_assignment)
        self.cluster_centers_ = X[chosen_centers]
        self.centers = chosen_centers



from typing import Optional

import numpy as np
import scipy

from holisticai.utils.models.recommender._rsbase import RecommenderSystemBase
from holisticai.utils.transformers.bias import BMInprocessing as BMImp

from .algorithm import DebiasingLearningAlgorithm


class DebiasingLearningMF(BMImp, RecommenderSystemBase):
    """
    Debiasing Learning Matrix Factorization handles selection biases by adapting
    models and estimation techniques from causal inference. The strategy leads to
    unbiased performance estimators despite biased data, and to a matrix factorization
    method that provides substantially improved prediction performance on real-world data.

    References:
            [1] Schnabel, Tobias, et al. "Recommendations as treatments: Debiasing learning
            and evaluation." international conference on machine learning. PMLR, 2016.
    """

    def __init__(
        self,
        K: Optional[int] = 10,
        normalization: Optional[str] = "Vanilla",
        lamda: Optional[float] = 0.02,
        metric: Optional[str] = "mse",
        bias_mode: Optional[str] = None,
        clip_val: Optional[float] = -1,
        seed: Optional[int] = None,
        verbose: Optional[int] = 0,
    ):
        """
        Init Debiasing Learning Matrix Factorization

        Parameters
        ----------

        K : int
                Specifies the number of dimensions.

        normalization : str
                Strategy to normalize rating matrix. Avaiables are:
                - 'Vanilla',
        - 'SelfNormalized'
        - 'UserNormalized'
        - 'ItemNormalized'

        lamda : float
                Model parameter.

        metric: str
                Metric used as cost function.

        clip_val: float
                Propensity Clip Value

        seed: int
                Random Seed

        bias_mode: str
                Bias value using in the model:
                - "None": No bias
                - "Free": Use bias wihtout regularizer in the cost function.
                - "Regularized": Use bias with regularizer in the cost function.

        verbose : int
                If >0, will show progress percentage.
        """
        self.seed = seed
        self.normalization = normalization
        self.metric = metric
        self.K = K
        self.clip_val = clip_val
        self.lamda = lamda
        self.bias_mode = bias_mode
        self.verbose = verbose
        self.algorithm = DebiasingLearningAlgorithm(
            K=K,
            normalization=normalization,
            metric=metric,
            lamda=lamda,
            clip_val=clip_val,
            bias_mode=bias_mode,
            verbose=self.verbose,
        )
        # self.params = [lamda, K, clipVal, bias_mode]

    def fit(self, X: Optional[np.ndarray], propensities: Optional[np.ndarray] = None):
        """
        Fit model

        Parameters
        ----------

        X : matrix-like
                rating matrix, 0 means non-raked cases.

        propensities : matrix-like (optional)
                Propensity matrix
        """
        invP = None
        if propensities is not None:
            invP = np.reciprocal(propensities)
            invP = np.ma.array(
                invP, copy=False, mask=np.ma.getmask(X), fill_value=0, hard_mask=True
            )

        # Get starting params by SVD
        params0 = self._init_parameters(X)
        model_params = self.algorithm.train(X, invP, params0)
        self.pred = self.algorithm.predict(model_params, bias_mode=self.bias_mode)
        self.invP = invP
        return self

    def _init_parameters(self, partial_observations):
        averageObservedRating = np.ma.mean(partial_observations)
        completeRatings = np.ma.filled(
            partial_observations.astype(np.float), averageObservedRating
        )
        numUsers, numItems = np.shape(partial_observations)
        numUsers = completeRatings.shape[0]
        ncv = (min(numUsers, numItems) + self.K) // 2

        u, s, vt = scipy.sparse.linalg.svds(
            completeRatings,
            k=self.K,
            ncv=ncv,
            tol=1e-7,
            which="LM",
            v0=None,
            maxiter=2000,
            return_singular_vectors=True,
        )
        P = u
        Q = np.transpose(np.multiply(vt, s[:, None]))
        userBias = np.zeros(numUsers)
        itemBias = np.zeros(numItems)
        params0 = (P, Q, userBias, itemBias, averageObservedRating)
        return params0



import numpy as np

from holisticai.bias.mitigation.inprocessing.commons import Logging


class MFLogger:
    def __init__(self, tau, eps, steps, verbose):
        self.verbose = verbose
        log_params = [("iteration", int), ("accuracy", float), ("gamma", float)]
        self.total_iterations = np.ceil(tau / eps) // steps
        self.logger = Logging(
            log_params=log_params,
            total_iterations=self.total_iterations,
            logger_format="iteration",
        )

    def update(self, it, max_acc, gamma):
        if self.verbose:
            self.logger.update(it, max_acc, gamma)



from calendar import c
from typing import Optional

import numpy as np

from holisticai.utils.transformers.bias import BMPostprocessing as BMPost
from holisticai.utils.transformers.bias import SensitiveGroups

from .algorithm import MulticlassBalancerAlgorithm


class LPDebiaserMulticlass(BMPost):
    """
    Linear Programmin Debiaser is a postprocessing algorithms designed to debias pretrained classifiers.
    The algorithm use constraints such as Equalized Odds and Equalized Opportunity.
    This technique extends LPDebiaserBinary for multiclass classification.
    References:
        Putzel, Preston, and Scott Lee. "Blackbox Post-Processing for Multiclass Fairness."
        arXiv preprint arXiv:2201.04461 (2022).
    """

    CONSTRAINT = ["EqualizedOdds", "EqualizedOpportunity"]
    OBJ_FUN = ["macro", "micro"]

    def __init__(
        self,
        constraint: Optional["CONSTRAINT"] = "EqualizedOdds",
        loss: Optional["OBJ_FUN"] = "macro",
    ):
        """
        Parameters
        ----------
        constraint : str
            Strategy used to evalute the cost function  The available contraints  are:
            [
                "EqualizedOdds",
                "EqualizedOpportunity"
            ]

        loss : str
            The loss function to optimize:
            [
                "macro",
                "micro"
            ],
            default "macro"
        """
        self.constraint = constraint
        self.loss = loss
        self.sens_groups = SensitiveGroups()

    def fit(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """

        Description
        ----------
        Compute parameters for Linear Programming Debiaser.
        For binary classification y_pred or y_proba can be used.
        For Multiclass classification only y_pred must be used.
        Parameters
        ----------
        y_true : array-like
            Target vector
        y_pred : array-like
            Predicted label vector (num_examples,).
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        Returns
        -------
        Self
        """
        params = self._load_data(
            y_true=y_true, y_pred=y_pred, group_a=group_a, group_b=group_b
        )

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_true = params["y_true"]
        y_pred = params["y_pred"]

        sensitive_features = np.stack([group_a, group_b], axis=1)
        p_attr = self.sens_groups.fit_transform(
            sensitive_features, convert_numeric=True
        )

        constraints_catalog, objective_catalog = self._get_catalogs()

        constraint = constraints_catalog[self.constraint]()

        objective = objective_catalog[self.loss]()

        self.algorithm = MulticlassBalancerAlgorithm(
            constraint=constraint, objective=objective
        )

        self.algorithm.fit(y_true=y_true, y_pred=y_pred, p_attr=p_attr)
        return self

    def transform(
        self,
        y_pred: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Apply transform function to predictions and likelihoods

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (nb_examlpes,)
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        dictionnary with new predictions
        """

        params = self._load_data(y_pred=y_pred, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_pred = params["y_pred"]

        sensitive_features = np.stack([group_a, group_b], axis=1)
        p_attr = self.sens_groups.transform(sensitive_features, convert_numeric=True)
        new_y_pred = self.algorithm.predict(y_pred=y_pred, p_attr=p_attr)
        return {"y_pred": new_y_pred}

    def fit_transform(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit and transform

        Parameters
        ----------
        y_true : array-like
            Target vector
        y_pred : array-like
            Predicted vector (nb_examlpes,)
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        dictionnary with new predictions
        """
        return self.fit(
            y_true=y_true,
            y_pred=y_pred,
            group_a=group_a,
            group_b=group_b,
        ).transform(y_pred=y_pred, group_a=group_a, group_b=group_b)

    def _get_catalogs(self):
        from .constraints import (
            EqualizedOdds,
            EqualizedOpportunity,
            MacroLosses,
            MicroLosses,
            Strict,
        )

        cons_cat = {
            "EqualizedOdds": EqualizedOdds,
            "EqualizedOpportunity": EqualizedOpportunity,
            "Strict": Strict,
        }

        obj_cat = {"macro": MacroLosses, "micro": MicroLosses}

        return cons_cat, obj_cat



from collections import defaultdict

import numpy as np

from holisticai.utils.transformers.bias import SensitiveGroups

from ._vanilla import VanillaFairletDecomposition

EPSILON = 0.0001


class TreeNode:
    def __init__(self):
        self.children = []

    def set_cluster(self, cluster):
        self.cluster = cluster

    def add_child(self, child):
        self.children.append(child)

    def populate_colors(self, colors):
        "Populate auxiliary lists of red and blue points for each node, bottom-up"
        self.reds = []
        self.blues = []
        if len(self.children) == 0:
            # Leaf
            for i in self.cluster:
                if colors[i] == 0:
                    self.reds.append(i)
                else:
                    self.blues.append(i)
        else:
            # Not a leaf
            for child in self.children:
                child.populate_colors(colors)
                self.reds.extend(child.reds)
                self.blues.extend(child.blues)


class ScalableFairletDecomposition(VanillaFairletDecomposition):
    def __init__(self, p, q):
        self.fairlets = []
        self.fairlet_centers = []
        assert p <= q, "Please use balance parameters in the correct order"
        self.p = p
        self.q = q
        self.sensgroups = SensitiveGroups()

    def _decompose(self, node, dataset, donelist, depth):
        p = self.p
        q = self.q

        # Leaf
        if len(node.children) == 0:
            node.reds = [i for i in node.reds if donelist[i] == 0]
            node.blues = [i for i in node.blues if donelist[i] == 0]
            assert self.balanced(
                p, q, len(node.reds), len(node.blues)
            ), "Reached unbalanced leaf"
            return super()._decompose(node.blues, node.reds, dataset)

        # Preprocess children nodes to get rid of points that have already been clustered
        for child in node.children:
            child.reds = [i for i in child.reds if donelist[i] == 0]
            child.blues = [i for i in child.blues if donelist[i] == 0]

        R = [len(child.reds) for child in node.children]
        B = [len(child.blues) for child in node.children]

        if sum(R) == 0 or sum(B) == 0:
            assert (
                sum(R) == 0 and sum(B) == 0
            ), "One color class became empty for this node while the other did not"
            return 0

        NR = 0
        NB = 0

        # Phase 1: Add must-remove nodes
        for i in range(len(node.children)):
            if R[i] >= B[i]:
                must_remove_red = max(0, R[i] - int(np.floor(B[i] * q * 1.0 / p)))
                R[i] -= must_remove_red
                NR += must_remove_red
            else:
                must_remove_blue = max(0, B[i] - int(np.floor(R[i] * q * 1.0 / p)))
                B[i] -= must_remove_blue
                NB += must_remove_blue

        # Calculate how many points need to be added to smaller class until balance
        if NR >= NB:
            # Number of missing blues in (NR,NB)
            missing = max(0, int(np.ceil(NR * p * 1.0 / q)) - NB)
        else:
            # Number of missing reds in (NR,NB)
            missing = max(0, int(np.ceil(NB * p * 1.0 / q)) - NR)

        # Phase 2: Add may-remove nodes until (NR,NB) is balanced or until no more such nodes
        for i in range(len(node.children)):
            if missing == 0:
                assert self.balanced(p, q, NR, NB), "Something went wrong"
                break
            if NR >= NB:
                may_remove_blue = B[i] - int(np.ceil(R[i] * p * 1.0 / q))
                remove_blue = min(may_remove_blue, missing)
                B[i] -= remove_blue
                NB += remove_blue
                missing -= remove_blue
            else:
                may_remove_red = R[i] - int(np.ceil(B[i] * p * 1.0 / q))
                remove_red = min(may_remove_red, missing)
                R[i] -= remove_red
                NR += remove_red
                missing -= remove_red

        # Phase 3: Add unsatuated fairlets until (NR,NB) is balanced
        for i in range(len(node.children)):
            if self.balanced(p, q, NR, NB):
                break
            if R[i] >= B[i]:
                num_saturated_fairlets = int(R[i] / q)
                excess_red = R[i] - q * num_saturated_fairlets
                excess_blue = B[i] - p * num_saturated_fairlets
            else:
                num_saturated_fairlets = int(B[i] / q)
                excess_red = R[i] - p * num_saturated_fairlets
                excess_blue = B[i] - q * num_saturated_fairlets
            R[i] -= excess_red
            NR += excess_red
            B[i] -= excess_blue
            NB += excess_blue

        assert self.balanced(p, q, NR, NB), "Constructed node sets are unbalanced"

        reds = []
        blues = []
        for i in range(len(node.children)):
            for j in node.children[i].reds[R[i] :]:
                reds.append(j)
                donelist[j] = 1
            for j in node.children[i].blues[B[i] :]:
                blues.append(j)
                donelist[j] = 1

        assert len(reds) == NR and len(blues) == NB, "Something went horribly wrong"
        return super()._decompose(blues, reds, dataset) + sum(
            [
                self._decompose(child, dataset, donelist, depth + 1)
                for child in node.children
            ]
        )

    def fit_transform(self, dataset, group_a, group_b):
        sensitive_groups = np.c_[group_a, group_b]
        colors = self.sensgroups.fit_transform(sensitive_groups, convert_numeric=True)
        root = build_quadtree(dataset)
        "Main fairlet clustering function, returns cost wrt original metric (not tree metric)"
        p = self.p
        q = self.q
        root.populate_colors(colors)
        assert self.balanced(
            p, q, len(root.reds), len(root.blues)
        ), "Dataset is unbalanced"
        root.populate_colors(colors)
        donelist = [0] * dataset.shape[0]
        cost = self._decompose(root, dataset, donelist, 0)
        return self.fairlets, self.fairlet_centers, cost


def build_quadtree(dataset, max_levels=0, random_shift=True):
    "If max_levels=0 there no level limit, quadtree will partition until all clusters are singletons"
    dimension = dataset.shape[1]
    lower = np.amin(dataset, axis=0)
    upper = np.amax(dataset, axis=0)

    shift = np.zeros(dimension)
    if random_shift:
        for d in range(dimension):
            spread = upper[d] - lower[d]
            shift[d] = np.random.uniform(0, spread)
            upper[d] += spread

    return build_quadtree_aux(
        dataset, range(dataset.shape[0]), lower, upper, max_levels, shift
    )


def build_quadtree_aux(dataset, cluster, lower, upper, max_levels, shift):
    """
    "lower" is the "bottom-left" (in all dimensions) corner of current hypercube
    "upper" is the "upper-right" (in all dimensions) corner of current hypercube
    """

    dimension = dataset.shape[1]
    cell_too_small = True
    for i in range(dimension):
        if upper[i] - lower[i] > EPSILON:
            cell_too_small = False

    node = TreeNode()
    if max_levels == 1 or len(cluster) <= 1 or cell_too_small:
        # Leaf
        node.set_cluster(cluster)
        return node

    # Non-leaf
    midpoint = 0.5 * (lower + upper)
    subclusters = defaultdict(list)
    for i in cluster:
        subclusters[
            tuple([dataset[i, d] + shift[d] <= midpoint[d] for d in range(dimension)])
        ].append(i)
    for edge, subcluster in subclusters.items():
        sub_lower = np.zeros(dimension)
        sub_upper = np.zeros(dimension)
        for d in range(dimension):
            if edge[d]:
                sub_lower[d] = lower[d]
                sub_upper[d] = midpoint[d]
            else:
                sub_lower[d] = midpoint[d]
                sub_upper[d] = upper[d]
        node.add_child(
            build_quadtree_aux(
                dataset, subcluster, sub_lower, sub_upper, max_levels - 1, shift
            )
        )
    return node



# Base Imports
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

# utils
from ...utils import get_colors
from ...utils._validation import _regression_checks
from ..metrics import confusion_matrix


def group_pie_plot(y_feat, ax=None, size=None, title=None):
    """Plot a pie chart showing proportions of groups in a feature.

    Parameters
    ----------
    y_feat : pandas series or numpy array
        Feature of interest
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # checks
    if isinstance(y_feat, pd.Series):
        value_counts = y_feat.value_counts()
        labels = value_counts.index.tolist()
    elif isinstance(y_feat, np.ndarray):
        y_feat = pd.Series(y_feat)
        value_counts = y_feat.value_counts()
        labels = value_counts.index.tolist()

    else:
        raise TypeError("input is not a numpy array or pandas series")

    # calculations
    n_b = np.sum(value_counts / np.sum(value_counts) > 0.02)
    n_nb = len(value_counts) - n_b
    if n_nb > 0 and n_b > 0:
        labels = list(labels[:n_b]) + ["Others"]
        value_counts = list(value_counts[:n_b]) + [np.sum(value_counts[n_b:])]

    # setup
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("Group proportions")

    # chart
    colors = get_colors(len(value_counts), extended_colors=True)
    hai_palette = sns.color_palette(colors)
    ax.pie(value_counts, labels=labels, colors=hai_palette, autopct="%.0f%%")

    # return
    return ax


def _group_confusion_matrices(group_a, group_b, y_pred, y_true, size=None, title=None):
    """Group confusion matrices comparison.

    Plots three heatmaps showing confusion matrices for
    two groups and the combination of those groups.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Prediction vector (categorical)
    y_true : array-like
        True values (categorical)
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # check and coerce
    group_a, group_b, y_pred, y_true, _ = _regression_checks(
        group_a, group_b, y_pred, y_true, None
    )

    # Calculating confusion matrices
    a_indices = group_a == 1
    conf_mat_a = confusion_matrix(y_pred[a_indices], y_true[a_indices])
    conf_mat_a = conf_mat_a.to_numpy()
    # normalise on columns
    conf_mat_a = conf_mat_a / conf_mat_a.sum(axis=0)[np.newaxis, :]
    b_indices = group_b == 1
    conf_mat_b = confusion_matrix(y_pred[b_indices], y_true[b_indices])
    conf_mat_b = conf_mat_b.to_numpy()
    # normalise on columns
    conf_mat_b = conf_mat_b / conf_mat_b.sum(axis=0)[np.newaxis, :]
    conf_mat_combi = confusion_matrix(y_pred, y_true)
    conf_mat_combi = conf_mat_combi.to_numpy()
    # normalise on columns
    conf_mat_combi = conf_mat_combi / conf_mat_combi.sum(axis=0)[np.newaxis, :]

    # Setup
    sns.set()
    colors = get_colors(100, extended_colors=True)
    hai_palette = sns.color_palette(colors)
    fig, ax = plt.subplots(1, 3, figsize=size)
    if title is not None:
        fig.suptitle(title)
    else:
        fig.suptitle("Confusion matrices")

    # charting
    sns.heatmap(
        conf_mat_a,
        annot=True,
        cmap=hai_palette,
        vmin=0,
        vmax=1,
        ax=ax[0],
        cbar=0,
        square=True,
    )
    sns.heatmap(
        conf_mat_b,
        annot=True,
        cmap=hai_palette,
        vmin=0,
        vmax=1,
        ax=ax[1],
        cbar=0,
        square=True,
    )
    sns.heatmap(
        conf_mat_combi,
        annot=True,
        cmap=hai_palette,
        vmin=0,
        vmax=1,
        ax=ax[2],
        cbar=1,
        square=True,
    )
    ax[0].set_title("group_a")
    ax[1].set_title("group_b")
    ax[2].set_title("Both groups")

    # return
    return ax


def distribution_plot(y_feat, p_attr=None, ax=None, size=None, title=None):
    """Plot the Kernel Density Estimation (KDE) distribution of the input data.
    If the protected attribute is included, the distributions are plotted for each
    group independently.

    Parameters
    ----------
    y_feat : pandas series or numpy array (regression)
        Feature of interest
    p_attr : pandas series or numpy array (categorical)
        Protected attribute
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # setup
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("KDE distribution plot")

    # charting
    colors = get_colors(len(np.unique(p_attr)), extended_colors=True)
    hai_palette = sns.color_palette(colors)
    ax = sns.kdeplot(
        x=y_feat, hue=p_attr, fill=True, palette=hai_palette, common_norm=False, ax=ax
    )

    # return
    return ax


def histogram_plot(y_feat, p_attr=None, ax=None, size=None, title=None):
    """Plot the histogram for categorical data.
    If the protected attribute is included, the histograms are plotted for each
    group independently.

    Parameters
    ----------
    y_feat : pandas series or numpy array (categorical)
        Feature of interest
    p_attr : pandas series or numpy array (categorical)
        Protected attribute
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
    matplotlib ax
    """
    # setup
    sns.set()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)
        if title is not None:
            fig.suptitle(title)
        else:
            fig.suptitle("Histogram Plot")

    colors = get_colors(len(np.unique(p_attr)))
    hai_palette = sns.color_palette(colors)

    if p_attr is None:
        hai_color = hai_palette[0]
        hai_palette = None
    else:
        hai_color = None

    # charting
    sns.histplot(
        x=y_feat,
        hue=p_attr,
        stat="probability",
        fill=True,
        palette=hai_palette,
        color=hai_color,
        common_norm=False,
        ax=ax,
    )
    _, labels = plt.xticks()
    plt.setp(labels, rotation=45)

    # return
    return ax



import itertools

import numpy as np
import scipy.sparse.csgraph


def fair_k_center_exact(dmat, p_attr, nr_centers_per_group, given_centers):
    """
    Description
    -----------
    Exhaustive search to exactly solve the fair k-center problem
    Obs: only works for small problem instances.

    Parameters
    ----------
    dmat : matrix-like
        distance matrix of size nxn
    p_attr :  array-like
        integer-vector of length n with entries in 0,...,m-1, where m is the number of groups
    nr_centers_per_group: array-like
        integer-vector of length m with entries in 0,...,k and sum over entries equaling k
    given_centers : array-like
        integer-vector with entries in 0,...,n-1

    Return
    -------
    (optimal centers, clustering, optimal fair k-center cost)
    """

    n = dmat.shape[0]
    m = nr_centers_per_group.size
    k = np.sum(nr_centers_per_group)

    cost = np.inf
    best_choice = []

    for mmm in itertools.combinations(np.arange(n), k):

        cluster_centers = np.array(mmm)

        curr_nr_clusters_per_sex = np.zeros(m)
        for ell in np.arange(m):
            curr_nr_clusters_per_sex[ell] = np.sum(p_attr[cluster_centers] == ell)

        if sum(curr_nr_clusters_per_sex == nr_centers_per_group) == m:
            curr_cost = np.amax(
                np.amin(
                    dmat[
                        np.ix_(
                            np.hstack((cluster_centers, given_centers)), np.arange(n)
                        )
                    ],
                    axis=0,
                )
            )
        else:
            curr_cost = np.inf

        if curr_cost < cost:
            cost = curr_cost
            best_choice = cluster_centers.copy()

    clustering = np.array(
        [
            np.argmin(dmat[ell, np.hstack((best_choice, given_centers))])
            for ell in np.arange(n)
        ]
    )

    return best_choice, clustering, cost


def k_center_greedy_with_given_centers(dmat, k, given_centers):
    """
    Description
    -----------
        Implementation of Algorithm 1.

    Parameters
    ----------
    dmat : matrix-like
        distance matrix of size nxn
    k : int
        integer smaller than n
    given_centers : array-like
        integer-vector with entries in 0,...,n-1

    Return
    ------
        approx. optimal centers
    """

    n = dmat.shape[0]

    if k == 0:
        cluster_centers = np.array([], dtype=int)
    else:
        if given_centers.size == 0:
            cluster_centers = np.random.choice(n, 1, replace=False)
            kk = 1
        else:
            cluster_centers = given_centers
            kk = 0

        distance_to_closest = np.amin(
            dmat[np.ix_(cluster_centers, np.arange(n))], axis=0
        )
        while kk < k:
            temp = np.argmax(distance_to_closest)
            cluster_centers = np.append(cluster_centers, temp)
            distance_to_closest = np.amin(
                np.vstack((distance_to_closest, dmat[temp, :])), axis=0
            )
            kk += 1

        cluster_centers = cluster_centers[given_centers.size :]

    return cluster_centers


def fair_k_center_APPROX(dmat, p_attr, nr_centers_per_group, given_centers):
    """
    Description
    -----------
        Implementation of Algorithm 4.

    Parameters
    ----------
    dmat : array-like
        distance matrix of size nxn
    p_attr :  array-like
        integer-vector of length n with entries in 0,...,m-1, where m is the number of groups
    nr_centers_per_group : array-like
        integer-vector of length m with entries in 0,...,k and sum over entries equaling k
    given_centers : array-like
        integer-vector with entries in 0,...,n-1

    Return
    ------
        approx. optimal centers
    """

    n = dmat.shape[0]
    m = nr_centers_per_group.size
    k = np.sum(nr_centers_per_group)

    cluster_centersTE = k_center_greedy_with_given_centers(dmat, k, given_centers)

    CURRENT_nr_clusters_per_sex = np.zeros(m, dtype=int)
    for ell in np.arange(k):
        CURRENT_nr_clusters_per_sex[p_attr[cluster_centersTE[ell]]] += 1

    partition = np.array(
        [
            np.argmin(dmat[ell, np.hstack((cluster_centersTE, given_centers))])
            for ell in np.arange(n)
        ]
    )
    selection = partition < k
    selected_partition = partition[selection]
    selected_p_attr = p_attr[selection]
    selected_index = np.array(
        [
            np.where(np.arange(n)[selection] == cluster_centersTE[ell])[0][0]
            for ell in np.arange(k)
        ]
    )
    G, centersTE = swapping_graph(
        selected_partition, selected_index, selected_p_attr, nr_centers_per_group
    )
    cluster_centersTE = np.arange(n)[partition < k][centersTE]

    if G.size == 0:
        cluster_centers = cluster_centersTE
    else:
        new_data_set = np.array([], dtype=int)
        new_given_centersT = np.array([], dtype=int)
        for ell in np.arange(k):
            if np.isin(p_attr[cluster_centersTE[ell]], G):
                new_data_set = np.hstack((new_data_set, np.where(partition == ell)[0]))
            else:
                new_given_centersT = np.hstack(
                    (new_given_centersT, cluster_centersTE[ell])
                )
        new_given_centers = np.hstack((new_given_centersT, given_centers))
        p_attr_new = p_attr[new_data_set]
        p_attr_newT = np.zeros(new_data_set.size, dtype=int)
        cc = 0
        for ell in G:
            p_attr_newT[p_attr_new == ell] = cc
            cc += 1
        new_data_set = np.hstack((new_data_set, new_given_centers))
        p_attr_newT = np.hstack(
            (p_attr_newT, np.zeros(new_given_centers.size, dtype=int))
        )

        cluster_centers_rek = fair_k_center_APPROX(
            dmat[np.ix_(new_data_set, new_data_set)],
            p_attr_newT,
            nr_centers_per_group[G],
            np.arange(new_data_set.size - new_given_centers.size, new_data_set.size),
        )

        new_given_centersT_additional = np.array([], dtype=int)
        for ell in np.setdiff1d(np.arange(m), G):
            if np.sum(p_attr[new_given_centersT] == ell) < nr_centers_per_group[ell]:
                toadd = nr_centers_per_group[ell] - np.sum(
                    p_attr[new_given_centersT] == ell
                )
                toadd_pot = np.setdiff1d(np.where(p_attr == ell)[0], new_given_centersT)
                if toadd_pot.size > toadd:
                    new_given_centersT_additional = np.hstack(
                        (new_given_centersT_additional, toadd_pot[0:toadd])
                    )
                else:
                    new_given_centersT_additional = np.hstack(
                        (new_given_centersT_additional, toadd_pot)
                    )

        cluster_centers = np.hstack(
            (
                new_given_centersT,
                new_given_centersT_additional,
                new_data_set[cluster_centers_rek],
            )
        )

    return cluster_centers


def swapping_graph(partition, centers, p_attr, nr_centers_per_group):
    """
    Description
    -----------
        Implementation of Algorithm 3.

    Parameters
    ----------
        partition : array-like
            integer-vector of length n with entries in 0 ... k-1
        centers : array-like
            integer-vector of length k with entries in 0 ... n-1
        p_attr : array-like
            integer-vector of length n with entries in 0 ... m-1
        nr_centers_per_group : array-like
            integer-vector of length m with entries in 0,...,k and sum over entries equaling k

    Return
    ------
        (G, swapped centers)
    """

    n = partition.size
    m = nr_centers_per_group.size
    k = centers.size

    CURRENT_nr_clusters_per_sex = np.zeros(m, dtype=int)
    for ell in np.arange(k):
        CURRENT_nr_clusters_per_sex[p_attr[centers[ell]]] += 1

    sex_of_assigned_center = p_attr[centers[partition]]
    Adja = np.zeros((m, m))
    for ell in np.arange(n):
        Adja[sex_of_assigned_center[ell], p_attr[ell]] = 1

    dmat_gr, predec = scipy.sparse.csgraph.shortest_path(
        Adja, directed=True, return_predecessors=True
    )

    is_there_a_path = 0
    for ell in np.arange(m):
        for zzz in np.arange(m):
            if (CURRENT_nr_clusters_per_sex[ell] > nr_centers_per_group[ell]) and (
                CURRENT_nr_clusters_per_sex[zzz] < nr_centers_per_group[zzz]
            ):
                if dmat_gr[ell, zzz] != np.inf:
                    path = np.array([zzz])
                    while path[0] != ell:
                        path = np.hstack((predec[ell, path[0]], path))
                    is_there_a_path = 1
                    break
        if is_there_a_path == 1:
            break

    while is_there_a_path:

        for hhh in np.arange(path.size - 1):
            for ell in np.arange(n):
                if (p_attr[ell] == path[hhh + 1]) and (
                    sex_of_assigned_center[ell] == path[hhh]
                ):
                    centers[partition[ell]] = ell
                    sex_of_assigned_center[partition == partition[ell]] = p_attr[ell]
                    break
        CURRENT_nr_clusters_per_sex[path[0]] -= 1
        CURRENT_nr_clusters_per_sex[path[-1]] += 1

        Adja = np.zeros((m, m))
        for ell in np.arange(n):
            Adja[sex_of_assigned_center[ell], p_attr[ell]] = 1

        dmat_gr, predec = scipy.sparse.csgraph.shortest_path(
            Adja, directed=True, return_predecessors=True
        )

        is_there_a_path = 0
        for ell in np.arange(m):
            for zzz in np.arange(m):
                if (CURRENT_nr_clusters_per_sex[ell] > nr_centers_per_group[ell]) and (
                    CURRENT_nr_clusters_per_sex[zzz] < nr_centers_per_group[zzz]
                ):
                    if dmat_gr[ell, zzz] != np.inf:
                        path = np.array([zzz])
                        while path[0] != ell:
                            path = np.hstack((predec[ell, path[0]], path))
                        is_there_a_path = 1
                        break
            if is_there_a_path == 1:
                break

    if sum(CURRENT_nr_clusters_per_sex == nr_centers_per_group) == m:
        return np.array([]), centers
    else:

        G = np.where(CURRENT_nr_clusters_per_sex > nr_centers_per_group)[0]
        for ell in np.arange(m):
            for zzz in np.arange(m):
                if ((dmat_gr[ell, zzz] != np.inf) and np.isin(ell, G)) and (
                    not np.isin(zzz, G)
                ):
                    G = np.hstack((G, zzz))

        return G, centers


def heuristic_greedy_on_each_group(dmat, p_attr, nr_centers_per_group, given_centers):
    """
    Description
    -----------
        Implementation of Heuristic A as described in Section 5.3.

    Parameters
    ----------
    dmat : matrix-like
        distance matrix of size nxn
    p_attr : array-like
        integer-vector of length n with entries in 0,...,m-1, where m is the number of groups
    nr_centers_per_group : array-like
        integer-vector of length m with entries in 0,...,k and sum over entries equaling k
    given_centers : array-like
        integer-vector with entries in 0,...,n-1

    Return
    ------
        heuristically chosen centers
    """

    m = nr_centers_per_group.size

    cluster_centers = np.array([], dtype=int)

    for ell in np.arange(m):
        subgroup = np.where(p_attr == ell)[0]
        given_centers_subgroup = np.where(np.isin(subgroup, given_centers))[0]

        cent_subgroup = k_center_greedy_with_given_centers(
            dmat[np.ix_(subgroup, subgroup)],
            nr_centers_per_group[ell],
            given_centers_subgroup,
        )

        cluster_centers = np.hstack((cluster_centers, subgroup[cent_subgroup]))

    return cluster_centers


def heuristic_greedy_till_constraint_is_satisfied(
    dmat, p_attr, nr_centers_per_group, given_centers
):
    """
    Description
    -----------
        Implementation of Heuristic B as described in Section 5.3.

    Parameters
    ----------
    dmat : matrix-like
        distance matrix of size nxn
    p_attr : array-like
        integer-vector of length n with entries in 0,...,m-1, where m is the number of groups
    nr_centers_per_group : array-like
        integer-vector of length m with entries in 0,...,k and sum over entries equaling k
    given_centers : array-like
        integer-vector with entries in 0,...,n-1

    Return
    ------
        heuristically chosen centers

    """

    n = dmat.shape[0]
    m = nr_centers_per_group.size
    k = np.sum(nr_centers_per_group)

    current_nr_per_sex = np.zeros(m)

    if k == 0:
        cluster_centers = np.array([], dtype=int)
    else:
        if given_centers.size == 0:
            cluster_centers = np.random.choice(n, 1, replace=False)
            current_nr_per_sex[p_attr[cluster_centers]] += 1
            kk = 1
        else:
            cluster_centers = given_centers
            kk = 0

        distance_to_closest = np.amin(
            dmat[np.ix_(cluster_centers, np.arange(n))], axis=0
        )
        while kk < k:
            feasible_groups = np.where(current_nr_per_sex < nr_centers_per_group)[0]
            feasible_points = np.where(np.isin(p_attr, feasible_groups))[0]
            new_point = feasible_points[np.argmax(distance_to_closest[feasible_points])]
            current_nr_per_sex[p_attr[new_point]] += 1
            cluster_centers = np.append(cluster_centers, new_point)
            distance_to_closest = np.amin(
                np.vstack((distance_to_closest, dmat[new_point, :])), axis=0
            )
            kk += 1

        cluster_centers = cluster_centers[given_centers.size :]

    return cluster_centers



from typing import Optional, Union

import numpy as np
import pandas as pd

from holisticai.utils.transformers._transformer_base import BMTransformerBase


class BMPreprocessing(BMTransformerBase):
    """
    Base Preprocessing transformer
    """

    BM_NAME = "Preprocessing"

    def _load_data(self, **kargs):
        """Save postprocessing atributes and convert data to standard format parameters."""

        params = {}
        if "y_true" in kargs:
            y_true = np.array(kargs.get("y_true")).ravel()
            params.update({"y_true": y_true})

        params_to_numpy_format = ["group_a", "group_b"]
        for param_name in params_to_numpy_format:
            if param_name in kargs:
                params.update({param_name: self._to_numpy(kargs, param_name)})

        if "X" in kargs:
            params.update({"X": self._to_numpy(kargs, "X", ravel=False)})

        if ("sample_weight" in kargs) and (kargs["sample_weight"] is not None):
            params.update({"sample_weight": self._to_numpy(kargs, "sample_weight")})

        elif "y_true" in locals():
            params.update({"sample_weight": np.ones_like(y_true).astype(np.float64)})

        elif "X" in kargs:
            params.update(
                {"sample_weight": np.ones(len(params["X"])).astype(np.float64)}
            )

        return params



import errno
import os
import os.path as osp
import timeit

import numpy as np
import scipy.io as sio
from scipy import sparse
from sklearn.neighbors import NearestNeighbors


def mkdir_if_missing(directory):
    if not osp.exists(directory):
        try:
            os.makedirs(directory)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise


def normalizefea(X):
    """
    L2 normalize
    """
    feanorm = np.maximum(1e-14, np.sum(X**2, axis=1))
    X_out = X / (feanorm[:, None] ** 0.5)
    return X_out


def get_V_jl(x, l, N, K):
    x = x.squeeze()
    temp = np.zeros((N, K))
    index_cluster = l[x]
    temp[(x, index_cluster)] = 1
    temp = temp.sum(0)
    return temp


def get_fair_accuracy(group_prob, groups_ids, l, K):
    N = len(groups_ids)
    V_j_list = np.array(
        [get_V_jl(groups_ids[f"{g}"], l, N, K) for g in group_prob.index]
    )

    balance = np.zeros(K)
    J = len(group_prob)
    for k in range(K):
        V_j_list_k = V_j_list[:, k].copy()
        balance_temp = np.tile(V_j_list_k, [J, 1])
        balance_temp = balance_temp.T / np.maximum(balance_temp, 1e-20)
        mask = np.ones(balance_temp.shape, dtype=bool)
        np.fill_diagonal(mask, 0)
        balance[k] = balance_temp[mask].min()

    return balance.min(), balance.mean()


def get_fair_accuracy_proportional(group_prob, groups_ids, l, K):
    N = len(groups_ids)
    V_j_list = np.array(
        [get_V_jl(groups_ids[f"{g}"], l, N, K) for g in group_prob.index]
    )
    clustered_uV = V_j_list / sum(V_j_list)
    fairness_error = np.zeros(K)
    u_V = np.array(group_prob)
    for k in range(K):
        fairness_error[k] = (
            -u_V * np.log(np.maximum(clustered_uV[:, k], 1e-20)) + u_V * np.log(u_V)
        ).sum()
    return fairness_error.sum()


def create_affinity(X, knn, scale=None, savepath=None, W_path=None):
    N, D = X.shape
    if W_path is not None:
        if W_path.endswith(".mat"):
            W = sio.loadmat(W_path)["W"]
        elif W_path.endswith(".npz"):
            W = sparse.load_npz(W_path)
    else:

        print("Compute Affinity ")
        start_time = timeit.default_timer()

        nbrs = NearestNeighbors(n_neighbors=knn).fit(X)
        dist, knnind = nbrs.kneighbors(X)

        row = np.repeat(range(N), knn - 1)
        col = knnind[:, 1:].flatten()
        if scale is None:
            data = np.ones(X.shape[0] * (knn - 1))
        elif scale is True:
            scale = np.median(dist[:, 1:])
            data = np.exp((-dist[:, 1:] ** 2) / (2 * scale**2)).flatten()
        else:
            data = np.exp((-dist[:, 1:] ** 2) / (2 * scale**2)).flatten()

        W = sparse.csc_matrix((data, (row, col)), shape=(N, N), dtype=np.float)
        W = (W + W.transpose(copy=True)) / 2
        elapsed = timeit.default_timer() - start_time
        print(elapsed)

        if isinstance(savepath, str):
            if savepath.endswith(".npz"):
                sparse.save_npz(savepath, W)
            elif savepath.endswith(".mat"):
                sio.savemat(savepath, {"W": W})

    return W



from typing import Optional, Tuple, Union

import numpy as np
from scipy.optimize import linprog
from sklearn.metrics import confusion_matrix

from holisticai.utils.transformers.bias import BMPostprocessing as BMPost


class EqualizedOdds(BMPost):
    """
    Equalized odds postprocessing use linear programming to find the probability with which change
    favorable labels (y=1) to unfavorable labels (y=0) in the output estimator to optimize equalized odds.

    References:
        Hardt, Moritz, Eric Price, and Nati Srebro. "Equality of opportunity in supervised learning."
        Advances in neural information processing systems 29 (2016).

        Pleiss, Geoff, et al. "On fairness and calibration."
        Advances in neural information processing systems 30 (2017).
    """

    SOLVERS = [
        "highs",
        "highs-ds",
        "highs-ipm",
        "interior-point",
        "revised simplex",
        "simplex",
    ]

    def __init__(self, solver: Optional[str] = "highs", seed: Optional[int] = 42):
        """
        Create a Equalized Odds Post-processing instance.

        Parameters
        ----------
        solver : str
            Algorithm name used to solve the standard form problem. Solver supported must depend of your scipy poackage version.
            for scipy 1.9.0 the solvers available are:
            ["highs", "highs-ds", "highs-ipm", "interior-point", "revised simplex", "simplex"]

        seed : int
            Random seed for repeatability.
        """
        self.solver = solver
        self.seed = seed

    def _objective_function_parameters_by_group(self, labels, predictions, group):
        # Compute Equalized Odds algorithm parameters for specific group

        labels = labels[group]
        predictions = predictions[group]

        prev = predictions
        flip_prev = 1 - predictions

        flip_predictions = 1 - predictions
        flip_labels = 1 - labels
        base_rate = np.mean(labels)

        tnr, fpr, fnr, tpr = confusion_matrix(
            labels, predictions, normalize="true"
        ).ravel()

        m_tn = np.logical_and(flip_predictions, flip_labels)
        m_fn = np.logical_and(flip_predictions, labels)
        m_fp = np.logical_and(predictions, flip_labels)
        m_tp = np.logical_and(predictions, labels)

        c = np.array([fpr - tpr, tnr - fnr])

        true_positive = (np.mean(prev * m_tp) - np.mean(flip_prev * m_tp)) / base_rate
        false_negative = (np.mean(flip_prev * m_fn) - np.mean(prev * m_fn)) / base_rate
        false_positive = (np.mean(prev * m_fp) - np.mean(flip_prev * m_fp)) / (
            1 - base_rate
        )
        true_negative = (np.mean(flip_prev * m_tn) - np.mean(prev * m_tn)) / (
            1 - base_rate
        )

        A_eq = np.array(
            [[true_positive, false_negative], [false_positive, true_negative]]
        )

        b_tp_fn = (np.mean(flip_prev * m_tp) + np.mean(prev * m_fn)) / base_rate
        b_fp_tn = (np.mean(flip_prev * m_fp) + np.mean(prev * m_tn)) / (1 - base_rate)

        b_eq = np.array([b_tp_fn, b_fp_tn])

        return {"c": c, "A_eq": A_eq, "b_eq": b_eq}

    def _build_objective_function(self, params_group_a, params_group_b):
        """
        Concat parameters from group a and b to create the objective function

        Parameters
        ----------
        params_group_a : dict
            dictionary of parameters with A_eq, b_eq, C keys.
        params_group_b : dict
            dictionary of parameters with A_eq, b_eq, C keys.

        Returns
        -------
        Tuple of objetive function parameters (A_eq, b_eq, C)
        """
        A_eq = np.concatenate([params_group_a["A_eq"], -params_group_b["A_eq"]], axis=1)
        b_eq = params_group_b["b_eq"] - params_group_a["b_eq"]
        C = np.concatenate([params_group_a["c"], params_group_b["c"]], axis=0)
        return A_eq, b_eq, C

    def _adjust_fairness(self, predictions, likelihoods, n2p, p2p):
        """
        Invert predictions and likelihoods with some probabilities

        Description
        ----------
        Use n2p and (1 - p2p) to invert output labels

        Parameters
        ----------
        predictions : numpy array
            binary predicted vector
        likelihoods : numpy array
            predicted score vector
        n2p : numpy array
            probability vector
        p2p : numpy array
            probability vector

        Returns
        -------
        numpy array
            new prediction vector with some inverted values

        numpy array
            new score vector with some inverted values
        """

        fair_predictions = predictions.copy()
        pos_indices = np.where(predictions == 1)[0]
        neg_indices = np.where(predictions == 0)[0]

        np.random.shuffle(pos_indices)
        np.random.shuffle(neg_indices)

        n2p_indices = neg_indices[: int(len(neg_indices) * n2p)]
        fair_predictions[n2p_indices] = 1

        p2n_indices = pos_indices[: int(len(pos_indices) * (1 - p2p))]
        fair_predictions[p2n_indices] = 0

        fair_likelihoods = likelihoods.copy()
        fair_likelihoods[n2p_indices] = 1 - fair_likelihoods[n2p_indices]
        fair_likelihoods[p2n_indices] = 1 - fair_likelihoods[p2n_indices]

        return fair_predictions, fair_likelihoods

    def fit(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ) -> "EqualizedOdds":
        """
        Compute parameters for equalizing odds.

        Description
        ----------
        Build parameters for the objetive function and call the solver to find the algorithm parameters.

        Parameters
        ----------
        y_true : array-like
            Target vector (nb_examlpes,)
        y_pred : array-like
            Predicted vector (nb_examlpes,)
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        Self
        """
        if self.seed is not None:
            np.random.seed(self.seed)

        params = self._load_data(
            y_true=y_true, y_pred=y_pred, group_a=group_a, group_b=group_b
        )

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_true = params["y_true"]
        y_pred = params["y_pred"]

        parameters_group_a = self._objective_function_parameters_by_group(
            y_true, y_pred, group_a
        )
        parameters_group_b = self._objective_function_parameters_by_group(
            y_true, y_pred, group_b
        )
        A_eq, b_eq, C = self._build_objective_function(
            parameters_group_a, parameters_group_b
        )

        A_ub = np.array(
            [
                [1, 0, 0, 0],
                [-1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, -1, 0, 0],
                [0, 0, 1, 0],
                [0, 0, -1, 0],
                [0, 0, 0, 1],
                [0, 0, 0, -1],
            ],
            dtype=np.float64,
        )

        b_ub = np.array([1, 0, 1, 0, 1, 0, 1, 0], dtype=np.float64)

        # Linear program
        self.model_params = linprog(
            C, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq=b_eq, method=self.solver
        )

        return self

    def transform(self, y_pred: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Apply transform function to predictions and likelihoods

        Description
        ----------
        Use the fitted probability to change the output label and invert the likelihood

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (nb_examlpes,)
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        A dictionary with two keys, y_pred and y_score, which refers to the predicted labels and their probabilities, respectively.
        """
        params = self._load_data(y_pred=y_pred, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_pred = params["y_pred"]
        likelihoods = y_pred.copy().astype(np.float64)

        sp2p, sn2p, op2p, on2p = self.model_params.x

        predictions_a = y_pred[group_a]
        predictions_b = y_pred[group_b]
        likelihoods_a = likelihoods[group_a]
        likelihoods_b = likelihoods[group_b]

        # Randomly flip labels according to the probabilities in model_params
        fair_predictions_a, fair_likelihoods_a = self._adjust_fairness(
            predictions_a, likelihoods_a, sn2p, sp2p
        )
        fair_predictions_b, fair_likelihoods_b = self._adjust_fairness(
            predictions_b, likelihoods_b, on2p, op2p
        )

        # Mutated, fairer dataset with new labels4
        new_y_pred = y_pred.copy()
        new_y_pred[group_a] = fair_predictions_a
        new_y_pred[group_b] = fair_predictions_b

        new_y_score = likelihoods.copy()
        new_y_score[group_a] = fair_likelihoods_a
        new_y_score[group_b] = fair_likelihoods_b

        return {"y_pred": new_y_pred, "y_score": new_y_score}
